/*
 * This is a driver for the Aachen DBCC CAMAC controller.
 * Author: Ole Streicher <ole@ifh.de>
 * The driver can be used under the terms of the GNU public license.
 * However, I do not give any warranty that it works or even is harmless.
 * Be carefully! This driver will destroy your hardware and send your hard 
 * disk to my mail address! You are warned!
 */

/*
 * There are two possibilities to use this driver:
 * 1. directly compiled-in
 *   Copy the file camac.c to linux/drivers/char/, camac.h to 
 *   linux/include/linux/ and install the patch camac.patch. Then recompile 
 *   your kernel, answering the question for a CAMAC driver with "yes".
 *   The Makefile target "builtin" will do exactly this.
 *   At boot time, you are able to give the base address of the interface 
 *   card with the option "camac=0x320"
 * 2. as a loadable module
 *   I recommend this option because it is easier to install and to handle.
 *   Type "make" to compile the module, and "make install" to install it.
 *   The command "insmod camac.o" will install the driver, the option 
 *   "base=0x320" will change the Base address, "major=25" will change the 
 *   major number. Use "rmmod camac.o" to remove the module.
 */

/*
 * History:
 *  0.1 initial development version (ole 1.6.95)
 *  0.2 loadable modules and setup function added (ole 6.6.95)
 *  0.3 tested with a real crate (ole 7.6.95)
 *  0.4 start including AC_WAIT_LAM (ole 11.6.95)
 *  0.4.1 bug in ac_wait_lam(): we need to set the crate every time we
 *        read after scheduling (ole 14.6.95)
 *  0.4.2 change the memory allocation for block r/w to the common kmalloc()
 *        (ole 16.6.95)
 *  0.4.3 added the accamac_enabled variable for builtin driver; now we can 
 *        disable the driver at LILO cmd line (ole 18.4.95)
 *  0.4.4 added the AC_START_CYCLE to the ioctl Camac commands (9.8.95)
 *        ignore NULL pointer in the ioctl commands (9.8.95)
 *  0.4.5 ignore NULL pointer in the r/w commands (10.8.95)
 *  0.4.6 AC_WAIT_LAM changed to AC_SET_LAM, new AC_WAIT_LAM waits 
 *         now directly for a LAM (10.8.95)
 *  0.4.7 added AC_START_CYCLE to the AC_WAIT_LAM ioctl command
 *        add a check the the user can really read/write the buffer (16.8.95)
 * 
 */

/*
 * Todo:
 *  - make a pre-compiles module for linux-1.2.10
 *  - test the functions:
 *    * write_blk..........        (far in the future)
 *  - write a library
 *  - think if would be better to make some ioctl options (AC_WAIT_LAM, 
 *    AC_BLOCKMODE) not globally for a given crate, but locally for each
 *    address (and subaddress? and function?)
 *  - will we need a select() function?
 *  - if we get an interrupt-capable board: use the interrupt
 *  - write a tk/tcl interface
 */

#include <linux/camac.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <asm/io.h>
#include <asm/segment.h>
#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h> 
static char kernel_version[] = UTS_RELEASE;
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#define CAMAC_VERSION "0.4.7"
#define STARTUPSTRING "DBCC CAMAC driver "CAMAC_VERSION" by Ole Streicher\n"

/*
 * The functions for the virtual file system
 */

static int lseek_accamac(struct inode *, struct file *, off_t, int );
static int read_accamac(struct inode *, struct file *, char *, int );
static int write_accamac(struct inode *, struct file *, char *, int );
static int ioctl_accamac(struct inode *, struct file *, unsigned, 
			 unsigned long);
static int open_accamac(struct inode *, struct file *);
static void close_accamac(struct inode *, struct file *);

/*
 * The functions for loading and exiting the driver
 */

#ifdef MODULES
int init_module(void);
void cleanup_module(void);
#else
unsigned long accamac_init(unsigned long);
void accamac_setup(char *, int *);
static int accamac_enabled = 1;
#endif

/*
 * The following functions are only internal used; they contain the
 * I/O code
 */

static int accamac_test(void);              /* tests the actual crate */
static void accamac_set(unsigned);          /* sets the crate nr */
static int accamac_test_pcboard(void);      /* test the PC interface */
static void accamac_allow_broadcast(void);  /* allow broadcasting */
static void accamac_deny_broadcast(void);   /* deny broadcasting */
static int accamac_status(void);            /* return X and Q status */
static int accamac_ctrl(int);               /* CAMAC control functions */
static void accamac_czs2(void);             /* make a c-z-s2 cycle */
static int accamac_wait_lam(int);           /* wait for LAM */
static void accamac_load_naf(int);          /* 1st part of CAMAC cycle */
static int accamac_read_16(void);           /* 2nd part: read 16 bit */
static int accamac_read_24(void);           /* 2nd part: read 24 bit */
static void accamac_read_blk(short *, int); /* 2nd part: read in block mode */
static void accamac_write_16(int);          /* 2nd part: write 16 bit */
static void accamac_write_24(int);          /* 2nd part: write 24 bit */
static void accamac_write_blk(short *, int);/* 2nd part: write in block mode */

/* 
 * here come some static variables
 * the start address is tunable: insmod camac.o base=0x320 for loadable 
 * modules and camac=0x320 for built-in driver
 * Tunable for modules: insmod camac major=25
 */

static unsigned long base = AC_BASE_ADDRESS;
#define BaseAddress base
#define CtrlInAddress (base + 0x00)
#define CtrlOutAddress (base + 0x00)
#define DataOutAddress (base + 0x02)
#define DataInAddress (base + 0x04)
static int major = AC_MAJOR;

static int nDeviceOpen[AC_MAX_CRATES];
static int BlockMode[AC_MAX_CRATES];
static int TimeOut[AC_MAX_CRATES];
static int currentDevice;

/*
 * Here is the file_operation structure
 */

struct file_operations accamac_fops = {
  lseek_accamac,
  read_accamac,
  write_accamac,
  NULL,           /* readdir */
  NULL,           /* select */
  ioctl_accamac,
  NULL,           /* mmap */
  open_accamac,
  close_accamac,
};


/*
 * The function lseek_accamac() is used with the NAF macro as argument;
 * just origin 0 (from start) is allowed
 */

static int 
lseek_accamac(struct inode * inode, struct file * file, off_t offset, int orig)
{
  if (orig == 0) {
    file->f_pos = offset;
    return file->f_pos;
  }
  else
    return -EINVAL;
} 

/*
 * The function read_accamac() reads in single or block mode depending from
 * the value of BlockMode[]
 * In single mode, if one reads 1 or 2 bytes, a 16-bit-cycle is started
 * if one reads 3 or 4 bytes, a 24-bit-cycle is started
 * in block mode, We will only read an even number of bytes
 */

static int 
read_accamac(struct inode * inode, struct file * file , char * buf, int count)
{

  int nwords = count/2;
  unsigned int crate = MINOR(inode->i_rdev);
  int result;

  if (crate > AC_MAX_CRATES)
    return -ENODEV;

  if (nwords > AC_MAX_WORDS)
    return -EINVAL;

  if (nwords == 0) 
    return 0;

  /*
   * Check that we can really write to the buffer
   */

  if ((result = verify_area(VERIFY_WRITE,buf,count)) && (buf != NULL))
    return result;

  /*
   * If the function code is >= 16, it is a write operation
   * in this case we dont want to read anything
   */
  
  if (F_NAF(file->f_pos) >= 16)
    return 0;

  /*
   * look if the block transfer mode is on 
   */

  if (BlockMode[crate] == AC_BLOCK_MODE) {

    short *temp = kmalloc(nwords, GFP_USER);

    accamac_set(crate);
    if (accamac_wait_lam(file->f_pos)) return -EIO;
    accamac_load_naf(file->f_pos);
    accamac_read_blk(temp, nwords);
    if (buf != NULL)
      memcpy_tofs(buf, temp, 2*nwords);
    kfree(temp);
    return 2*nwords;
  } 
  else { /* if single mode */

    int temp;

    switch (count) {
    case 1:
      accamac_set(crate);
      if (accamac_wait_lam(file->f_pos)) return -EIO;
      accamac_load_naf(file->f_pos | AC_START_CYCLE);
      temp = accamac_read_16();
      if (buf != NULL)
       	put_fs_byte(temp, buf);
      return count;
    case 2:
      accamac_set(crate);
      if (accamac_wait_lam(file->f_pos)) return -EIO;
      accamac_load_naf(file->f_pos | AC_START_CYCLE);
      temp = accamac_read_16();
      if (buf != NULL)
       	put_fs_word(temp, buf);
      return count;
    case 3:
      accamac_set(crate);
      if (accamac_wait_lam(file->f_pos)) return -EIO;
      accamac_load_naf(file->f_pos | AC_START_CYCLE);
      temp = accamac_read_24();
      if (buf != NULL)
       	memcpy_tofs(buf, &temp, 3);
      return count;
    case 4:
      accamac_set(crate);
      if (accamac_wait_lam(file->f_pos)) return -EIO;
      accamac_load_naf(file->f_pos | AC_START_CYCLE);
      temp = accamac_read_24();
      if (buf != NULL)
        put_fs_long(temp, buf);
      return count;
    default:
      return -EINVAL;
    } /* switch count */
  } /* if ! blockmode */

}

/*
 * The function write_accamac() writes in single or block mode depending from
 * the value of BlockMode[]
 * In single mode, if one writes 1 or 2 bytes, a 16-bit-cycle is started
 * if one writes 3 or 4 bytes, a 24-bit-cycle is started
 * in block mode, We will only write an even number of bytes
 */

static int 
write_accamac(struct inode * inode, struct file * file , char * buf, int count)
{

  int nwords = count/2;
  unsigned int crate = MINOR(inode->i_rdev);
  int result;

  if (crate > AC_MAX_CRATES)
    return -ENODEV;

  if (nwords > AC_MAX_WORDS)
    return -EINVAL;

  if (nwords == 0) 
    return 0;

  /*
   * Check that we can really read from the buffer
   */

  if ((result = verify_area(VERIFY_READ,buf,count)))
    return result;
     
  /*
   * If the function code is < 16, it is a read operation
   * in this case we dont want to write anything
   */

  if (F_NAF(file->f_pos) < 16)
    return 0;

  /*
   * look if the block transfer mode is on 
   */
  
  if (BlockMode[crate] == AC_BLOCK_MODE) {

    short *temp = kmalloc(nwords, GFP_USER);

    memcpy_fromfs(temp, buf, 2*nwords);
    accamac_set(crate);
    accamac_load_naf(file->f_pos);
    accamac_write_blk(temp, nwords);
    kfree(temp);
    return 2*nwords;
  } 
  else { /* if single mode */

    int temp;

    switch (count) {
    case 1:
      temp = get_fs_byte(buf);
      accamac_set(crate);
      accamac_load_naf(file->f_pos);
      accamac_write_16(temp);
      return count;
    case 2:
      temp = get_fs_word(buf);
      accamac_set(crate);
      accamac_load_naf(file->f_pos);
      accamac_write_16(temp);
      return count;
    case 3:
      memcpy_fromfs(&temp, buf, 3);
      accamac_set(crate);
      accamac_load_naf(file->f_pos);
      accamac_write_24(temp);
      return count;
    case 4:
      temp = get_fs_long(buf);
      accamac_set(crate);
      accamac_load_naf(file->f_pos);
      accamac_write_24(temp);
      return count;
    default:
      return -EINVAL;
    } /* switch count */
  } /* if ! blockmode */

}


/*
 * The function ioctl_accamac() controls the behaviour of the driver
 */

static int 
ioctl_accamac(struct inode * inode, struct file * file, 
	      unsigned cmd, unsigned long arg)
{

  unsigned long *ptr = (unsigned long *)arg;
  unsigned int crate = MINOR(inode->i_rdev);
  unsigned long temp;
  int result;

  if (crate > AC_MAX_CRATES)
    return -ENODEV;

  switch (cmd) {
  case AC_C_Z_S2:
    accamac_set(crate);
    accamac_czs2();
    return 0;
  case AC_TEST_CRATE:
    if (crate == AC_BROADCAST_CRATE) return -ENOIOCTLCMD;
    accamac_set(crate);
    temp = accamac_test();
    if (ptr != NULL) {
      if ((result = verify_area(VERIFY_WRITE,ptr,sizeof(long))))
	return result;
      memcpy_tofs(ptr, &temp, sizeof(long));
    }
    return 0;
  case AC_STATUS:
    accamac_set(crate);
    temp = accamac_status();
    if (ptr != NULL) {
      if ((result = verify_area(VERIFY_WRITE,ptr,sizeof(long))))
	return result;
      memcpy_tofs(ptr, &temp, sizeof(long));
    }
    return 0;
  case AC_BLOCK_MODE:
    BlockMode[crate] = AC_BLOCK_MODE;
    return 0;
  case AC_SINGLE_MODE:
    BlockMode[crate] = AC_SINGLE_MODE;
    return 0;
  case AC_ALLOW_BROADCAST:
    if (crate == AC_BROADCAST_CRATE) return -ENOIOCTLCMD;
    accamac_set(crate);
    accamac_allow_broadcast();
    return 0;
  case AC_DENY_BROADCAST:
    if (crate == AC_BROADCAST_CRATE) return -ENOIOCTLCMD;
    accamac_set(crate);
    accamac_deny_broadcast();
    return 0;
  case AC_TEST_LAM:
  case AC_CLEAR_1:    
  case AC_CLEAR_LAM:
  case AC_CLEAR_2:
  case AC_DISABLE:
  case AC_EXECUTE:
  case AC_ENABLE:
  case AC_TEST_STATUS:
    temp = accamac_ctrl(NAF(N_NAF(file->f_pos), A_NAF(file->f_pos),cmd) 
			| AC_START_CYCLE);
    if (ptr != NULL) {
      if ((result = verify_area(VERIFY_WRITE,ptr,sizeof(long))))
	return result;
      memcpy_tofs(ptr, &temp, sizeof(long));
    }
    return 0;
  case AC_SET_LAM:
    if ((result = verify_area(VERIFY_READ,ptr,sizeof(long))))
      return result;
    TimeOut[crate] = get_fs_long(arg);
    if (TimeOut[crate] > AC_MAX_TIMEOUT)
      TimeOut[crate] = AC_MAX_TIMEOUT;
    return 0;
  case AC_NOWAIT_LAM:
    TimeOut[crate] = -1;
    return 0;
  case AC_WAIT_LAM:
    if (accamac_wait_lam(file->f_pos)) return -EIO;
    return 0;
  default: 
    return -ENOIOCTLCMD;
  }

}

/*
 * The function open_accamac() tests the crate and opens it
 */

static int 
open_accamac(struct inode * inode, struct file * file)
{

  unsigned int crate = MINOR(inode->i_rdev);

  if (crate > AC_MAX_CRATES)
    return -ENODEV;

  if (nDeviceOpen[crate] != 0)
    return -EBUSY;

  if (crate == AC_BROADCAST_CRATE) {        /* open the broadcasting device */
    MOD_INC_USE_COUNT;
    nDeviceOpen[crate]++;
    BlockMode[crate] = AC_SINGLE_MODE;
    TimeOut[crate] = -1;
    return 0;
  }
    
  accamac_set(crate);
  if (! accamac_test())
    return -ENODEV;

  BlockMode[crate] = AC_SINGLE_MODE;
  TimeOut[crate] = -1;          /* no time out -- direct poll */
  accamac_deny_broadcast();

  MOD_INC_USE_COUNT;
  nDeviceOpen[crate]++;

  return 0;

}


/*
 * The function close_accamac() closes the device
 */

void 
close_accamac(struct inode * inode, struct file * file)
{

  unsigned int crate = MINOR(inode->i_rdev);

  if (crate > AC_MAX_CRATES)
    return;

  if (crate == AC_BROADCAST_CRATE) { /* close the broadcasting crate */
    MOD_DEC_USE_COUNT;
    nDeviceOpen[crate] = 0;
    return;
  }

  accamac_set(crate);
  accamac_allow_broadcast();
  MOD_DEC_USE_COUNT;
  nDeviceOpen[crate] = 0;
  return;

}


/*
 * The function init_module() is used by starting the module
 * The function accamac_init() is used by Linux startup if the driver
 * is directly compiled in.
 */
#ifdef MODULE
#define INIT_FAIL -EIO
#define INIT_SUCCESS 0
#else
#define INIT_FAIL mem_start
#define INIT_SUCCESS mem_start
#endif

#ifdef MODULE
int 
init_module(void)
#else
unsigned long 
accamac_init(unsigned long mem_start)
#endif
{
  int crate;
  int count = 0;

#ifndef MODULES
  if (!accamac_enabled)
    return INIT_FAIL;
#endif

  if (register_chrdev(major, AC_NAME, &accamac_fops)) {
    printk(" unable to get major %i for " AC_NAME " device\n",major);
    return INIT_FAIL;
  }

  printk(STARTUPSTRING);

  if (check_region(BaseAddress, AC_BASE_EXTEND)) {
    printk(" unable to access i/o address 0x%lx\n",BaseAddress);
    unregister_chrdev(major, AC_NAME);
    return INIT_FAIL;
  }

  if (! accamac_test_pcboard()) {
    printk(" interface card not found at address 0x%lx\n",BaseAddress);
    unregister_chrdev(major, AC_NAME);
    return INIT_FAIL;
  }

  printk(" interface card found at address 0x%lx\n",BaseAddress);
  request_region(BaseAddress, AC_BASE_EXTEND, AC_NAME);

  currentDevice = -1; 
  for (crate = 0; crate < AC_MAX_CRATES; crate++) {
    
    nDeviceOpen[crate] = 0;

    if (crate == AC_BROADCAST_CRATE)
      continue;                    /* no check for broadcasting crate */

    accamac_set(crate);
    if (accamac_test()) {
      count++;
      printk(" crate /dev/camac%i available\n", crate);
    } /* if accamac_test() */
  } /* for crate */ 

  if (count == 0) 
    printk(" no crate found\n");

  return INIT_SUCCESS;

}

#ifdef MODULE
/*
 * cleanup_module() cleans up the module if it is not used
 * if compiled-in, there is ofcourse no cleanup possible
 */

void
cleanup_module(void)
{

  if (MOD_IN_USE) {
    printk(AC_NAME ": device busy, module not removed\n");
    return;
  }
  
  release_region(BaseAddress, AC_BASE_EXTEND);

  if (unregister_chrdev(major, AC_NAME) != 0)
    printk(AC_NAME ": cleanup_module failed\n");
  else
    printk(AC_NAME ": module removed\n");
}

#endif /* def MODULE */

#ifndef MODULE
/*
 * The setup takes only one (optional) parameter: the base address
 * type "camac=0x320" at LILO command line
 * for a loadable module, this can be done via "insmod camac.o base=0x320"
 */
void accamac_setup(char *str, int *ints)
{
  if (!strcmp(str, "off")) {
    accamac_enabled = 0;
    return;
  }

  if (ints[0] > 0)
    BaseAddress=ints[1];
  accamac_enabled = 1;
}
#endif /* ifndef MODULE */

/*
 * Here starts the real I/O code. It was mostly stolen
 * from Thomas Mikolajski's ACCAMAC Turbo-Pascal library
 * He gave the permission to use it here.
 */

/*
 * accamac_set() sets the current device if it is not already set
 */

static void
accamac_set(unsigned crate)
{
  
  if (currentDevice != crate) {

    outw(0, CtrlOutAddress);
    outw((0x10 | (crate & 0x0f)), DataOutAddress);

    currentDevice = crate;
  } /* if currentDevice != crate */
}


/*
 * accamac_test() tests the actual crate
 * and denies further broadcasting
 * 0 means: not found
 */

static int 
accamac_test(void)
{

  int cbus;
  int present = 1;

  accamac_deny_broadcast();
  outw_p(3, CtrlOutAddress); /* primary mode 3: read */
  cbus = inw_p(DataInAddress);
  if ((cbus & 0x04) == 0)
    present = 0;

  accamac_allow_broadcast();
  outw_p(3, CtrlOutAddress); /* primary mode 3: read */
  cbus = inw_p(DataInAddress);
  if ((cbus & 0x04) != 0)
    present = 0;

  outw_p(1, CtrlOutAddress); /* reset, clear crate */
  outw_p(6, DataOutAddress);
  outw_p(2, CtrlOutAddress); /* make c+z pulse */
  cbus = accamac_status();
  if (!AC_Q_FLAG(cbus)  || !AC_X_FLAG(cbus))
    present = 0;

  return present;
}


/*
 * accamac_test_pcboard() tests the PC interface card
 * 0 means: not found
 */

static int 
accamac_test_pcboard(void)
{
#if 0 /* is really a very ugly code!! */
   
  if (inw(DataInAddress) == 0xffff) /* quick and dirty -- may fail */
    return 0;
  else
    return 1;
#else
   return 1;
#endif
}

/*
 * accamac_deny_broadcast() excludes the current crate from the 
 * broadcast operations
 */
static void 
accamac_deny_broadcast(void)
{
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(7, DataOutAddress); /* submode 7 */
  outw_p(2, CtrlOutAddress); /* primary mode 2 */
  outw_p(4, DataOutAddress); /* submode 4: broadcasting off */
}

/*
 * accamac_allow_broadcast() allows  broadcast operations
 * at the current crate
 */

static void 
accamac_allow_broadcast(void)
{
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(7, DataOutAddress); /* submode 7 */
  outw_p(2, CtrlOutAddress); /* primary mode 2 */
  outw_p(0, DataOutAddress); /* submode 4: broadcasting on */
}


/*
 * accamac_status() reads out the current status of the C and Q flag
 */

static int 
accamac_status()
{
  return inw(CtrlInAddress);
}

/*
 * accamac_ctrl() invokes the several control commands
 */

static int 
accamac_ctrl(int naf)
{
  accamac_load_naf(naf);
  return AC_Q_FLAG(accamac_status());
}

/*
 * accamac_czs2() makes a C-Z-S2 cycle on the CAMAC dataway
 */

static void 
accamac_czs2(void)
{
  accamac_load_naf(0); 
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(6, DataOutAddress); /* submode 6 */
  outw_p(2, CtrlOutAddress); /* primary mode 2: exec CAMAC cycle with c&z */
}

/*
 * The timeout function we need for the Look-at-me waiting.
 * (see below)
 */

static void 
accamac_mark_timeout(unsigned long timed_out)
{
  *(int *)timed_out = 1;
}

/*
 * accamac_wait_lam() waits for the LAM-state if the timeout is >0
 * currently it is only a dummy, until I know how to make an processor-speed
 * independent timeout. If I get an interrup-capable interface, this 
 * function is the right point to add the interrupt code
 * currently, this function uses this ugly polling mode
 * The timeout should be given in milliseconds.
 * It returns zero on success and -EIO on error.
 */

static int accamac_wait_lam(int naf)
{
  int q;
  int crate;
  volatile int timed_out = 0;
  struct timer_list 
    timer_camac = {NULL, NULL, TimeOut[currentDevice],
		      (int)&timed_out, accamac_mark_timeout};

  /*
   * if the timer<0, dont look for LAM, just return.
   */

  if (TimeOut[currentDevice] <= 0)
    return 0;

  /*
   * replace the Function code with the look-at-me code
   */

  naf = NAF(N_NAF(naf), A_NAF(naf), AC_TEST_LAM) | AC_START_CYCLE;

  /*
   * Polling loop: q holds the Look-at-me flag.
   */

  q = accamac_ctrl(naf);
  add_timer(&timer_camac);

  /*
   * In the loop, we will allow a re-scheduling, so that other
   * processes may use another crate and change the current crate.
   * So, we must store here the current crate locally, and restore it
   * if nessessary.
   */

  crate = currentDevice;

  while ((!timed_out) && (q == 0)) {
    if (need_resched) {
      schedule();
      accamac_set(crate);
    }
    q = accamac_ctrl(naf);
  }

  del_timer(&timer_camac);
  if (timed_out)
    printk("camac: Timeout waiting for LAM\n");
  
  if (timed_out)
    return -EIO; /* failure*/
  else
    return 0;    /* success */
}


/*
 * accamac_load_naf() is the first part of the CAMAC cycle
 * it writes the Device number (n), the subaddress (a) and the
 * function code (f) to the interface. The NAF code can be calculated 
 * with the macro NAF(n, a, f)
 */

static void 
accamac_load_naf(int naf)
{
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(0, DataOutAddress); /* submode 0 */
  outw(2, CtrlOutAddress); /* primary mode 2 */
  outw(naf, DataOutAddress); /* load naf */
}


/*
 * accamac_read_16() is the second part of the CAMAC cycle; reads 16 bit
 */

static int 
accamac_read_16(void)
{
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(3, DataOutAddress); /* submode 3 */
  outw(3, CtrlOutAddress); /* primary mode 3: read */

  return inw(DataInAddress); /* read data */
}

/*
 * accamac_read_24() is the second part of the CAMAC cycle; reads 24 bit
 */

static int 
accamac_read_24(void)
{
  int data;

  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(3, DataOutAddress); /* submode 3 */
  outw(3, CtrlOutAddress); /* primary mode 3: read */
  data = inw(DataInAddress); /* read lower 16 bit data */
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(4, DataOutAddress); /* submode 4 */
  outw(3, CtrlOutAddress); /* primary mode 3: read */

  return ((inw(DataInAddress) & 0xff) << 16) + (data & 0xffff);/* high 8 bit */
}

/*
 * accamac_read_blk() is the second part of the CAMAC cycle; 
 * reads in block mode
 */

static void
accamac_read_blk(short *data, int nwords)
{
  int i;
  
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(5, DataOutAddress); /* submode 5 */
  outw(3, CtrlOutAddress); /* primary mode 3: read */

  for (i = 0; i < nwords-1; i++)
    data[i] = inw(DataInAddress);       /* read the data (16 bit) */

  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(3, DataOutAddress); /* submode 3 */
  outw(3, CtrlOutAddress); /* primary mode 3: read */
  data[i] = inw(DataInAddress); /* read the last data word */

}


/*
 * accamac_write_16() is the second part of the CAMAC cycle; writes 16 bit
 */

static void
accamac_write_16(int data)
{
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(1, DataOutAddress); /* submode 1 */
  outw(2, CtrlOutAddress); /* primary mode 2: write */
  outw(0x0000, DataOutAddress); /* write high order bits: 0 */
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(2, DataOutAddress); /* submode 2 */
  outw(2, CtrlOutAddress); /* primary mode 2: write */
  outw((data & 0xffff), DataOutAddress); /* write low order bits */
}

/*
 * accamac_write_24() is the second part of the CAMAC cycle; writes 24 bit
 */

static void
accamac_write_24(int data)
{
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(1, DataOutAddress); /* submode 1 */
  outw(2, CtrlOutAddress); /* primary mode 2: write */
  outw(((data & 0xff0000) >> 16), DataOutAddress); /* write high order bits */
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(2, DataOutAddress); /* submode 2 */
  outw(2, CtrlOutAddress); /* primary mode 2: write */
  outw((data & 0xffff), DataOutAddress); /* write low order bits */
}

/*
 * accamac_write_blk() is the second part of the CAMAC cycle; 
 * writes in block mode
 */

static void
accamac_write_blk(short *data, int nwords)
{
  int i;
  
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(2, DataOutAddress); /* submode 2 */
  outw(2, CtrlOutAddress); /* primary mode 2: write */
  for (i = 0; i < nwords; i++)
    outw(data[i], DataOutAddress); /* write data bits */

  printk("accamac_write_blk(*data, %i)\n",nwords);
}

/*
 * Here this driver code ends. Good luck!
 *
 * Tschuessi
 *
 * Ole
 */


